import mwclient as mc

from .article import Article
from .lists.listcustomtracks import ListCustomTracks, Table as LCTTable, Entry as LCTEntry
from .lists.listretrosothergames import ListRetrosOtherGames, Table as LROGTable, Entry as LROGEntry

class Wiiki(object):
    
    def __init__(self, username, password, api):
        wiiki = mc.Site("wiki.tockdom.com", custom_headers = {"User-Agent": api})
        wiiki.login(username, password)
        self.wiiki = wiiki
        self.username = username
    
    def __repr__(self):
        return "mwclient Wiiki object:\nLogged in as {}.".format(self.username)
    
    def article(self, title):
        if title == "List of Custom Tracks":
            return ListCustomTracks(self.wiiki)
        elif title == "List of Retro Tracks from Non-Mario Kart Games":
            return ListRetrosOtherGames(self.wiiki)
        else:
            return Article(title, self.wiiki)
    
    def search(self, query, where = "title", namespace = 0):
        l = []
        for article in self.wiiki.search(query, namespace, where):
            l.append(article.get("title"))
        if not bool(l):
            return None
        elif len(l) == 1:
            return l[0]
        else:
            return l
    
    def approximate_title(self, title):
        result = self.search(title)
        if result == None:
            return None
        elif type(result) == str:
            return result
        else:
            length = float("inf")
            for article in result:
                if title not in article:
                    continue
                if len(article) < length:
                    approximation = article
                    length = len(approximation)
            return approximation
    
    def text(self, title):
        page = Article(title, self.wiiki)
        return page.text
    
    def edit(self, title, text, summary = "", minor = False):
        page = Article(title, self.wiiki)
        page.object.edit(text, summary, minor)
    
    def move(self, title, new_title, summary = "", redirect = False):
        page = Article(title, self.wiiki)
        page.object.move(new_title, reason = summary, no_redirect = not redirect)
    
    def articles_in_category(self, name):
        category = self.wiiki.categories[name]
        names = set()
        for article in category:
            names.add(article.name)
        return names
    
    def categories_of_article(self, name):
        article = Article(name, self.wiiki)
        categories = set()
        for category in article.object.categories():
            categories.add(category.name)
        return categories



class Link(object):
    
    def __init__(self, string):
        if string.startswith("[[") and "|" in string:
            self.link = string[2:string.find("|")]
            self.display = string[string.find("|") + 1:-2]
            self.brackets = 2
        elif string.startswith("[["):
            self.link = string[2:-2]
            self.display = None
            self.brackets = 2
        elif string.startswith("["):
            self.link = string[1:string.find(" ")]
            self.display = string[string.find(" ") + 1:-1]
            self.brackets = 1
        else:
            self.link = string
            self.display = None
            self.brackets = 0
    
    def __str__(self):
        if self.brackets == 2 and self.display is None:
            return "[[" + self.link + "]]"
        elif self.brackets == 2 and self.display:
            return "[[" + self.link + "|" + self.display + "]]"
        elif self.brackets == 1:
            return "[" + self.link + " " + self.display + "]"
        else:
            return self.link
    
    def __repr__(self):
        return str(self)


from ..article import Article
from ..sortstring import sort_string

class ListRetrosOtherGames(object):
    
    def __init__(self, wiiki):
        self.title = "List of Retro Tracks from Non-Mario Kart Games"
        text = Article(self.title, wiiki).text
        
        self.templates = text[:text.find("This is a") - 1].split("\n")
        self.overview = text[text.find("This is a"):text.find("{| class") - 1]
        self.tracks = Table(text[text.find("{| class"):text.find("|}\n\n{|") + len("|}\n")])
        self.arenas = Table(text[text.rfind("{| class"):text.rfind("|}") + 3])
        self.categories = text[text.rfind("\n\n[[") + len("\n\n[[") - 2:].split("\n")
    
    def __str__(self):
        string = "\n".join(self.templates)
        string += "\n" + str(self.overview)
        string += "\n" + str(self.tracks)
        string += "\n" + str(self.arenas)
        string += "\n" + "\n".join(self.categories) + "\n"
        return string
    
    def __repr__(self):
        return str(self.title) + " article"

class Table(object):
    
    def __init__(self, text):
        class_pos = text.find("|-")
        self.classification = text[:class_pos].split("\n")
        text = text[class_pos + 3:]
        self.entries = []
        
        i = 0
        while True:
            next_separator = "|-\n" if i % 2 != 0 else "|- class=alt\n"
            self_separator = "|-\n" if i % 2 == 0 else "|- class=alt\n"
            
            first_piece = text[:text.find("]]")]
            if "data-sort-value" in first_piece and "rowspan" in first_piece:
                raise NotImplementedError("this type does not exist in the table yet")
            elif "data-sort-value" in first_piece:
                column = text[2:text.find("\n|-")].split("\n| ")
                sort = column[0][column[0].find("\"") + 1:column[0].rfind("\"")]
                title = column[0][column[0].rfind("|") + 1:-2]
                prefix = column[0][column[0].find("[[") + 2:column[0].find(title) - 1]
                entry = Entry(title, prefix, column[1], column[2], column[3], column[4], 1, sort)
            elif "rowspan" in first_piece:
                # needs to be done
                rows = int(text[len("| rowspan=")])
                title = first_piece[first_piece.rfind("|") + 1:]
                prefix = first_piece[first_piece.find("[[") + 2:first_piece.find(title) - 1]
                text = text[text.find("\n") + 1:]
                second_piece = text[:text.find("\n")]
                if "data-sort-value" in second_piece:
                    game = second_piece[second_piece.find("data-sort-value"):]
                else:
                    game = second_piece[second_piece.find("''"):]
                text = text[text.find("\n") + 1:]
                authors = []
                firsts = []
                latests = []
                for x in range(rows):
                    if x < rows - 1:
                        column = text[2:text.find(self_separator) - 1].split("\n| ")
                    else:
                        column = text[2:text.find(next_separator) - 1].split("\n| ")
                    authors.append(column[0])
                    firsts.append(column[1])
                    latests.append(column[2])
                    if x < rows - 1:
                        text = text[text.find(self_separator) + len(self_separator):]
                entry = Entry(title, prefix, game, authors, firsts, latests, rows)
            else:
                column = text[2:text.find("\n|-")].split("\n| ")
                title = column[0][column[0].find("|") + 1:-2]
                prefix = column[0][2:column[0].find(title) - 1]
                entry = Entry(title, prefix, column[1], column[2], column[3], column[4])
            
            self.entries.append(entry)
            
            if text.find("|-") == -1:
                break
            
            text = text[text.find(next_separator) + len(next_separator):]
            i += 1
        
        self.entries[-1].latest = self.entries[-1].latest[:-3]
    
    def __str__(self):
        string = "\n".join(self.classification)
        for entry, x in zip(self.entries, range(len(self.entries))):
            if x % 2 == 0:
                string += "|-\n" + str(entry)
            else:
                string += "|- class=alt\n" + entry.__str__(True)
        string += "|}\n"
        return string
    
    def __repr__(self):
        return "Retro Tracks from Non-Mario Kart Games table with {} entries".format(len(self.entries))
    
    def __add__(self, other):
        if type(other) != Entry:
            raise TypeError("only objects of type 'Entry' are supported")
        result = Table(str(self))
        for entry in result.entries:
            if entry.title == other.title:
                if entry.rowspan > 1:
                    entry.author.append(other.author)
                    entry.first.append(other.first)
                    entry.latest.append(other.latest)
                else:
                    entry.author = [entry.author, other.author]
                    entry.first = [entry.first, other.first]
                    entry.latest = [entry.latest, other.latest]
                entry.rowspan += 1
                return result
        result.entries.append(other)
        result.sort()
        return result
    
    def update_release(self, track, date, author = None):
        for entry in self.entries:
            if entry.prefix + " " + entry.title == track:
                if entry.rowspan > 1:
                    position = entry.author.index(author)
                    entry.latest[position] = date
                else:
                    entry.latest = date
                break
    
    def sort(self):
        self.entries = sorted(self.entries, key = lambda entry:entry.sortstring)

class Entry(object):
    
    def __init__(self, title, prefix, game, author, first, latest, rowspan = 1, sort = None):
        self.title = title
        self.prefix = prefix
        self.game = game
        self.author = author
        self.first = first
        self.latest = latest
        self.rowspan = rowspan
        self.sort = sort if sort else title
        self.sortstring = sort_string(self.sort)
    
    def __str__(self, special_separator = False):
        if self.sort != self.title and self.rowspan > 1:
            raise NotImplementedError("this type does not exist in the table yet")
        elif self.sort != self.title:
            string = "| data-sort-value=\"" + str(self.sort) + "\"| "
            string += "[[" + str(self.prefix) + " " + str(self.title) + "|" + str(self.title) + "]]\n"
            string += "| " + str(self.game) + "\n"
            string += "| " + str(self.author) + "\n"
            string += "| " + str(self.first) + "\n"
            string += "| " + str(self.latest) + "\n"
        elif self.rowspan > 1:
            string = "| rowspan=" + str(self.rowspan) + "| "
            string += "[[" + str(self.prefix) + " " + str(self.title) + "|" + str(self.title) + "]]\n"
            string += "| rowspan=" + str(self.rowspan)
            if "data-sort-value" in self.game:
                string += " " + str(self.game) + "\n"
            else:
                string += "| " + str(self.game) + "\n"
            for x in range(self.rowspan):
                string += "| " + str(self.author[x]) + "\n"
                string += "| " + str(self.first[x]) + "\n"
                string += "| " + str(self.latest[x]) + "\n"
                if x < self.rowspan - 1:
                    if special_separator:
                        string += "|- class=alt\n"
                    else:
                        string += "|-\n"
        else:
            string = "| [[" + str(self.prefix) + " " + str(self.title) + "|" + str(self.title) + "]]\n"
            string += "| " + str(self.game) + "\n"
            string += "| " + str(self.author) + "\n"
            string += "| " + str(self.first) + "\n"
            string += "| " + str(self.latest) + "\n"
        
        return string
    
    def __repr__(self):
        representation = ""
        representation += "Title: " + str(self.prefix) + " " + str(self.title) + "\n"
        representation += "Game: " + str(self.game) + "\n"
        representation += "Author(s): " + str(self.author) + "\n"
        representation += "First: " + str(self.first) + "\n"
        representation += "Latest: " + str(self.latest)
        if self.rowspan > 1:
            representation += "\nRowspan of " + str(self.rowspan)
        if self.sort != self.title:
            representation += "\nData sort value: " + str(self.sort)
        return representation



import mwclient as mc

class Article(object):
    
    def __init__(self, title, wiiki):
        self.title = title
        self.object = mc.page.Page(wiiki, title)
        self.text = self.object.text()



from ..article import Article
from ..templates.link import Link
from ..sortstring import sort_string

class ListCustomTracks(object):
    
    def __init__(self, wiiki):
        self.title = "List of Custom Tracks"
        text = Article(self.title, wiiki).text
        
        self.templates = text[:text.find("== Overview ==") - 1].split("\n")
        self.overview = text[text.find("== Overview ==") + len("== Overview ==") + 1:text.find("== Track List ==") - 1]
        self.table = Table(text[text.find("== Track List ==") + len("== Track List ==") + 1:text.find("== Translations ==") - 1])
        self.translations = text[text.find("== Translations ==") + len("== Translations ==") + 1:text.rfind("\n\n[[") + 1]
        self.categories = text[text.rfind("\n\n[[") + len("\n\n[[") - 2:].split("\n")
    
    def __str__(self):
        string = "\n".join(self.templates)
        string += "\n== Overview ==\n" + str(self.overview)
        string += "\n== Track List ==\n" + str(self.table)
        string += "\n== Translations ==\n" + str(self.translations)
        string += "\n" + "\n".join(self.categories) + "\n"
        return string
    
    def __repr__(self):
        return str(self.title) + " article"

class Table(object):
    
    def __init__(self, text):
        class_pos = text.find("|-")
        self.classification = text[:class_pos].split("\n")
        text = text[class_pos + 3:]
        self.entries = []
        
        while True:
            first_piece = text[:text.find("]]")]
            if "data-sort-value" in first_piece and "rowspan" in first_piece:
                data_sort = first_piece[first_piece.find("\"") + 1:first_piece.rfind("\"")]
                text = text.replace(" data-sort-value=\"" + data_sort + "\"", "")
                rows = int(text[len("| rowspan=")])
                title = text[text.find("[["):text.find("]]") + 2]
                text = text[text.find("\n") + 1:]
                authors = []
                firsts = []
                latests = []
                for x in range(rows):
                    column = text[2:text.find("\n|")].split(" || ")
                    authors.append(column[0])
                    firsts.append(column[1])
                    latests.append(column[2])
                    if x < rows - 1:
                        text = text[text.find("|-") + 3:]
                entry = Entry(title, authors, firsts, latests, rows, data_sort)
            elif "data-sort-value" in first_piece:
                column = text[2:text.find("\n|")].split(" || ")
                data_sort = column[0][column[0].find("\"") + 1:column[0].rfind("\"")]
                title = column[0][column[0].find("[["):column[0].find("]]") + 2]
                entry = Entry(title, column[1], column[2], column[3], 1, data_sort)
            elif "rowspan" in first_piece:
                rows = int(text[len("| rowspan=")])
                title = text[text.find("[["):text.find("]]") + 2]
                text = text[text.find("\n") + 1:]
                authors = []
                firsts = []
                latests = []
                for x in range(rows):
                    column = text[2:text.find("\n")].split(" || ")
                    authors.append(column[0])
                    firsts.append(column[1])
                    latests.append(column[2])
                    if x < rows - 1:
                        text = text[text.find("|-") + 3:]
                entry = Entry(title, authors, firsts, latests, rows)
            else:
                column = text[2:text.find("\n|")].split(" || ")
                entry = Entry(column[0], column[1], column[2], column[3])
            
            self.entries.append(entry)
            
            if text.find("\n") == text[:-1].rfind("\n"):
                break
            
            text = text[text.find("|-") + 3:]
    
    def __str__(self):
        string = "\n".join(self.classification)
        for entry in self.entries:
            string += "|-\n" + str(entry)
        string += "|}\n"
        return string
    
    def __repr__(self):
        return "Custom Track Table with {} entries".format(len(self.entries))
    
    def __add__(self, other):
        if type(other) != Entry:
            raise TypeError("only objects of type 'Entry' are supported")
        result = Table(str(self))
        for entry in result.entries:
            if entry.title == other.title:
                if entry.rowspan > 1:
                    entry.author.append(other.author)
                    entry.first.append(other.first)
                    entry.latest.append(other.latest)
                else:
                    entry.author = [entry.author, other.author]
                    entry.first = [entry.first, other.first]
                    entry.latest = [entry.latest, other.latest]
                entry.rowspan += 1
                return result
        result.entries.append(other)
        result.sort()
        return result
    
    def update_release(self, title, date, author = None):
        for entry in self.entries:
            if entry.title == title:
                if entry.rowspan > 1:
                    position = entry.author.index(author)
                    entry.latest[position] = date
                else:
                    entry.latest = date
                break
    
    def sort(self):
        self.entries = sorted(self.entries, key = lambda entry:entry.sortstring)

class Entry(object):
    
    def __init__(self, title, author, first, latest, rowspan = 1, data_sort = None):
        self.title = Link(title)
        self.author = author
        self.first = first
        self.latest = latest
        self.rowspan = rowspan
        self.data_sort = data_sort if data_sort else self.title.link
        self.sortstring = sort_string(self.data_sort)
    
    def __str__(self):
        if self.data_sort != self.title.link and self.rowspan > 1:
            string = "| data-sort-value=\"" + str(self.data_sort) + "\""
            string += " rowspan=" + str(self.rowspan) + "| "
            string += str(self.title) + "\n"
            for x in range(self.rowspan):
                string += "| " + str(self.author[x]) + " || "
                string += str(self.first[x]) + " || "
                string += str(self.latest[x]) + "\n"
                if x < self.rowspan - 1:
                    string += "|-\n"
        elif self.data_sort != self.title.link:
            string = "| data-sort-value=\"" + str(self.data_sort) + "\"| "
            string += str(self.title) + " || "
            string += str(self.author) + " || "
            string += str(self.first) + " || "
            string += str(self.latest) + "\n"
        elif self.rowspan > 1:
            string = "| rowspan=" + str(self.rowspan) + "| "
            string += str(self.title) + "\n"
            for x in range(self.rowspan):
                string += "| " + str(self.author[x]) + " || "
                string += str(self.first[x]) + " || "
                string += str(self.latest[x]) + "\n"
                if x < self.rowspan - 1:
                    string += "|-\n"
        else:
            string = "| "
            string += str(self.title) + " || "
            string += str(self.author) + " || "
            string += str(self.first) + " || "
            string += str(self.latest) + "\n"
        
        return string
    
    def __repr__(self):
        representation = ""
        representation += "Title: " + str(self.title) + "\n"
        representation += "Author(s): " + str(self.author) + "\n"
        representation += "First: " + str(self.first) + "\n"
        representation += "Latest: " + str(self.latest)
        if self.rowspan > 1:
            representation += "\nRowspan of " + str(self.rowspan)
        if self.data_sort != self.title:
            representation += "\nData sort value: " + str(self.data_sort)
        return representation



import importlib as il

characters = il.resources.files("wiiki_editor").joinpath("sort.txt").read_text("utf-8")
characters = characters.split("\n")

def sort_string(string):
    sort_string = ""
    for letter in string.lower():
        sort_string += chr(characters.index(letter))
    return sort_string



class Status(object):
    
    def __init__(self, status, variable = None):
        """
        0. No
        1. Yes
        2. Maybe
        """
        if status not in range(3):
            raise ValueError("only status 0, 1 and 2 are accepted")
        self.status = status
        self.variable = variable
    
    def __str__(self):
        if self.status == 0 and self.variable:
            return "{{no|" + str(self.variable) + "}}"
        elif self.status == 1 and self.variable:
            return "{{yes|" + str(self.variable) + "}}"
        elif self.status == 2 and self.variable:
            return "{{maybe|" + str(self.variable) + "}}"
        elif self.status == 0:
            return "{{no}}"
        elif self.status == 1:
            return "{{yes}}"
        elif self.status == 2:
            return "{{maybe}}"
    
    def __repr__(self):
        return str(self)
